--
-- USC emulation for USC board
--
-- 2010-09-26 by JTA
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.isync_pkg.all;
use work.stepdir_gen_pkg.all;
use work.pll_pkg.all;
use work.rate_gen_pkg.all;
use work.epp_if_pkg.all;

entity BoB_Main is
port
(
  clk     : in std_logic;
  input   : in std_logic_vector(8 downto 1);
  output  : out std_logic_vector(12 downto 1);
  pcpin   : inout std_logic_vector(17 downto 1)
);
end BoB_Main;

architecture rtl of BoB_Main is

type AXIS_T is record
  reg_ratelatch : std_logic_vector(23 downto 0);
  reg_rate : std_logic_vector(23 downto 0);
  pulse : std_logic;
  fwd : std_logic;
  rev : std_logic;
  step : std_logic;
  dir : std_logic;
end record;

type AXIS_ARRAY_T is array(natural range <>) of AXIS_T;

--
-- Rate generator control register bits
--

constant RGCTRL_BIT_AX0_DIR : integer := 0;
constant RGCTRL_BIT_AX0_RUN : integer := 1;
constant RGCTRL_BIT_AX1_DIR : integer := 2;
constant RGCTRL_BIT_AX1_RUN : integer := 3;
constant RGCTRL_BIT_AX2_DIR : integer := 4;
constant RGCTRL_BIT_AX2_RUN : integer := 5;
constant RGCTRL_BIT_AX3_DIR : integer := 6;
constant RGCTRL_BIT_AX3_RUN : integer := 7;

--
-- Encoder control register bits
--

constant ENCCTRL_BIT_ENC0_LOAD : integer := 4;
constant ENCCTRL_BIT_ENC1_LOAD : integer := 5;
constant ENCCTRL_BIT_ENC2_LOAD : integer := 6;
constant ENCCTRL_BIT_ENC3_LOAD : integer := 7;

--
-- Card type and version
--

constant CARD_ID : std_logic_vector(7 downto 0) := x"41";

--
-- Signals
--

signal reset      : std_logic := '0';
signal clk100M    : std_logic;
signal pll_locked : std_logic;

signal inputs : std_logic_vector(8 downto 1); -- synchronized input pins

signal addr : std_logic_vector(7 downto 0);

signal reg_encoder0 : std_logic_vector(23 downto 0);
signal reg_encoder0_latch : std_logic_vector(23 downto 0);
signal reg_encoder1 : std_logic_vector(23 downto 0);
signal reg_encoder1_latch : std_logic_vector(23 downto 0);
signal reg_encoder2 : std_logic_vector(23 downto 0);
signal reg_encoder2_latch : std_logic_vector(23 downto 0);
signal reg_encoder3 : std_logic_vector(23 downto 0);
signal reg_encoder3_latch : std_logic_vector(23 downto 0);
signal reg_encctrl : std_logic_vector(7 downto 0);

signal reg_encass : std_logic_vector(23 downto 0);
signal reg_rgctrl : std_logic_vector(7 downto 0);
signal reg_dout   : std_logic_vector(7 downto 0);

signal din        : std_logic_vector(7 downto 0);
signal dout       : std_logic_vector(7 downto 0);
signal rd         : std_logic;
signal wr         : std_logic;
signal waitreq    : std_logic;

signal axis : AXIS_ARRAY_T(0 to 2);

signal epp_int : std_logic;

begin

-- PLL

syspll_inst : syspll port map
(
    inclk0 => clk,
    c0 => clk100m,
    locked => pll_locked
);

reset <= not pll_locked;

-- PC parallel port pin assignments

pcpin(10) <= epp_int;
pcpin(12) <= 'Z';
pcpin(13) <= 'Z';
pcpin(15) <= 'Z';
epp_int <= '0';

--
-- Outputs
--

output(1) <= axis(0).dir;
output(2) <= axis(0).step;
output(3) <= not reg_dout(0);

output(4) <= axis(1).dir;
output(5) <= axis(1).step;
output(6) <= not reg_dout(1);

output(7) <= axis(2).dir;
output(8) <= axis(2).step;
output(9) <= not reg_dout(2);

output(10) <= reg_dout(3);
output(11) <= reg_dout(4);
output(12) <= reg_dout(5);

--
-- Digital input synchronizers
--

inp_idsync: isync_slv
generic map
(
  SYNC_STAGES => 2,
  WIDTH => input'length
)
port map
(
  clk => clk100m,
  reset => reset,
  sin => input,
  sout => inputs
);

--
-- EPP IF
--

epp_interface: epp_if
generic map
(
  WAITCOUNT => 10
)
port map
(
  clk => clk100m,
  reset => reset,
  --
  epp_data        => pcpin(9 downto 2),
  epp_wait        => pcpin(11),
  epp_dir_in      => pcpin(1),
  epp_astrobeb_in => pcpin(17),
  epp_dstrobeb_in => pcpin(14),
  epp_reset_in    => pcpin(16),
  --
  addr              => addr,
  dout              => din,
  din               => dout,
  rd                => rd,
  wr                => wr,
  waitrequest       => waitreq,
  reset_out         => open,
  addr_autoinc_ena  => '1'
);

waitreq <= '0';

--
-- Data read MUX
--

dout <= reg_encoder0(7 downto 0)          when addr = x"00" else
        reg_encoder0_latch(15 downto 8)   when addr = x"01" else
        reg_encoder0_latch(23 downto 16)  when addr = x"02" else
        reg_encoder1(7 downto 0)          when addr = x"03" else
        reg_encoder1_latch(15 downto 8)   when addr = x"04" else
        reg_encoder1_latch(23 downto 16)  when addr = x"05" else
        reg_encoder2(7 downto 0)          when addr = x"06" else
        reg_encoder2_latch(15 downto 8)   when addr = x"07" else
        reg_encoder2_latch(23 downto 16)  when addr = x"08" else
        reg_encoder3(7 downto 0)          when addr = x"09" else
        reg_encoder3_latch(15 downto 8)   when addr = x"0A" else
        reg_encoder3_latch(23 downto 16)  when addr = x"0B" else
        x"00"                             when addr = x"0C" else
        inputs                            when addr = x"0D" else
        x"80"                             when addr = x"0E" else
        CARD_ID                           when addr = x"0F" else
        x"00";

--
-- Axis 0 rate generator
--

axis0_rg: rate_gen
generic map
(
  REGWIDTH => axis(0).reg_rate'length,
  CLKDIV => 10
)
port map
(
  clk => clk100m,
  reset => reset,
  rate => axis(0).reg_rate,
  enable => reg_rgctrl(RGCTRL_BIT_AX0_RUN),
  pout => axis(0).pulse
);

axis(0).fwd <= '1' when axis(0).pulse = '1' and reg_rgctrl(RGCTRL_BIT_AX0_DIR) = '1' else '0';
axis(0).rev <= '1' when axis(0).pulse = '1' and reg_rgctrl(RGCTRL_BIT_AX0_DIR) = '0' else '0';

axis0_sdgen: stepdir_gen
generic map
(
  DIR_SETUP => 205,
  STEP_WIDTH => 125
)
port map
(
  clk => clk100m,
  reset => reset,
  fwd_in => axis(0).fwd,
  rev_in => axis(0).rev,
  enable => reg_rgctrl(RGCTRL_BIT_AX0_RUN),
  step => axis(0).step,
  dir => axis(0).dir
);

--
-- Axis 1 rate generator
--

axis1_rg: rate_gen
generic map
(
  REGWIDTH => axis(1).reg_rate'length,
  CLKDIV => 10
)
port map
(
  clk => clk100m,
  reset => reset,
  rate => axis(1).reg_rate,
  enable => reg_rgctrl(RGCTRL_BIT_AX1_RUN),
  pout => axis(1).pulse
);

axis(1).fwd <= '1' when axis(1).pulse = '1' and reg_rgctrl(RGCTRL_BIT_AX1_DIR) = '1' else '0';
axis(1).rev <= '1' when axis(1).pulse = '1' and reg_rgctrl(RGCTRL_BIT_AX1_DIR) = '0' else '0';

axis1_sdgen: stepdir_gen
generic map
(
  DIR_SETUP => 205,
  STEP_WIDTH => 125
)
port map
(
  clk => clk100m,
  reset => reset,
  fwd_in => axis(1).fwd,
  rev_in => axis(1).rev,
  enable => reg_rgctrl(RGCTRL_BIT_AX1_RUN),
  step => axis(1).step,
  dir => axis(1).dir
);

--
-- Axis 2 rate generator
--

axis2_rg: rate_gen
generic map
(
  REGWIDTH => axis(2).reg_rate'length,
  CLKDIV => 10
)
port map
(
  clk => clk100m,
  reset => reset,
  rate => axis(2).reg_rate,
  enable => reg_rgctrl(RGCTRL_BIT_AX2_RUN),
  pout => axis(2).pulse
);

axis(2).fwd <= '1' when axis(2).pulse = '1' and reg_rgctrl(RGCTRL_BIT_AX2_DIR) = '1' else '0';
axis(2).rev <= '1' when axis(2).pulse = '1' and reg_rgctrl(RGCTRL_BIT_AX2_DIR) = '0' else '0';

axis2_sdgen: stepdir_gen
generic map
(
  DIR_SETUP => 205,
  STEP_WIDTH => 125
)
port map
(
  clk => clk100m,
  reset => reset,
  fwd_in => axis(2).fwd,
  rev_in => axis(2).rev,
  enable => reg_rgctrl(RGCTRL_BIT_AX2_RUN),
  step => axis(2).step,
  dir => axis(2).dir
);

--
-- Encoder assembly register
--

reg_encass_p: process(clk100m, reset)
begin
  if reset = '1' then
    reg_encass <= (others => '0');
  elsif rising_edge(clk100m) then
    if wr = '1' then
      case addr is
        when x"00" =>
          reg_encass(7 downto 0) <= din;
        when x"01" =>
          reg_encass(15 downto 8) <= din;
        when x"02" =>
          reg_encass(23 downto 16) <= din;
        when others =>
          null;
      end case;
    end if;
  end if;
end process;

--
-- Encoder control register
--

reg_encctrl_p: process(clk100m, reset)
begin
  if reset = '1' then
    reg_encctrl <= (others => '0');
  elsif rising_edge(clk100m) then
    if addr = x"03" then
      if wr = '1' then
        reg_encctrl <= din;
      end if;
    end if;
  end if;
end process;

--
-- Encoder 0 counter register
--

reg_encoder0_p: process(clk100m, reset)
begin
  if reset = '1' then
    reg_encoder0 <= (others => '0');
  elsif rising_edge(clk100m) then
    if reg_encctrl(ENCCTRL_BIT_ENC0_LOAD) = '1' then
      reg_encoder0 <= reg_encass;
    elsif axis(0).fwd = '1' and axis(0).rev = '0' then
      reg_encoder0 <= std_logic_vector(signed(reg_encoder0) + 1);
    elsif axis(0).fwd = '0' and axis(0).rev = '1' then
      reg_encoder0 <= std_logic_vector(signed(reg_encoder0) - 1);
    end if;
  end if;
end process;

reg_encoder0latch_p: process(clk100m, reset)
begin
  if reset = '1' then
    reg_encoder0_latch <= (others => '0');
  elsif rising_edge(clk100m) then
    if addr = x"00" and rd = '1' then
      reg_encoder0_latch <= reg_encoder0;
    end if;
  end if;
end process;

--
-- Encoder 1 counter register
--

reg_encoder1_p: process(clk100m, reset)
begin
  if reset = '1' then
    reg_encoder1 <= (others => '0');
  elsif rising_edge(clk100m) then
    if reg_encctrl(ENCCTRL_BIT_ENC1_LOAD) = '1' then
      reg_encoder1 <= reg_encass;
    elsif axis(1).fwd = '1' and axis(1).rev = '0' then
      reg_encoder1 <= std_logic_vector(signed(reg_encoder1) + 1);
    elsif axis(1).fwd = '0' and axis(1).rev = '1' then
      reg_encoder1 <= std_logic_vector(signed(reg_encoder1) - 1);
    end if;
  end if;
end process;

reg_encoder1latch_p: process(clk100m, reset)
begin
  if reset = '1' then
    reg_encoder1_latch <= (others => '0');
  elsif rising_edge(clk100m) then
    if addr = x"03" and rd = '1' then
      reg_encoder1_latch <= reg_encoder1;
    end if;
  end if;
end process;

--
-- Encoder 2 counter register
--

reg_encoder2_p: process(clk100m, reset)
begin
  if reset = '1' then
    reg_encoder2 <= (others => '0');
  elsif rising_edge(clk100m) then
    if reg_encctrl(ENCCTRL_BIT_ENC2_LOAD) = '1' then
      reg_encoder2 <= reg_encass;
    elsif axis(2).fwd = '1' and axis(2).rev = '0' then
      reg_encoder2 <= std_logic_vector(signed(reg_encoder2) + 1);
    elsif axis(2).fwd = '0' and axis(2).rev = '1' then
      reg_encoder2 <= std_logic_vector(signed(reg_encoder2) - 1);
    end if;
  end if;
end process;

reg_encoder2latch_p: process(clk100m, reset)
begin
  if reset = '1' then
    reg_encoder2_latch <= (others => '0');
  elsif rising_edge(clk100m) then
    if addr = x"06" and rd = '1' then
      reg_encoder2_latch <= reg_encoder2;
    end if;
  end if;
end process;

--
-- Encoder 3 counter register (to satisfy USC test program)
--

reg_encoder3_p: process(clk100m, reset)
begin
  if reset = '1' then
    reg_encoder3 <= (others => '0');
  elsif rising_edge(clk100m) then
    if reg_encctrl(ENCCTRL_BIT_ENC3_LOAD) = '1' then
      reg_encoder3 <= reg_encass;
    end if;
  end if;
end process;

reg_encoder3latch_p: process(clk100m, reset)
begin
  if reset = '1' then
    reg_encoder3_latch <= (others => '0');
  elsif rising_edge(clk100m) then
    if addr = x"09" and rd = '1' then
      reg_encoder3_latch <= reg_encoder3;
    end if;
  end if;
end process;

--
-- Rate generator 0 rate register
--

reg_ax0_rate_p: process(clk100m, reset)
begin
  if reset = '1' then
    axis(0).reg_rate <= (others => '0');
    axis(0).reg_ratelatch <= (others => '0');
  elsif rising_edge(clk100m) then
    axis(0).reg_ratelatch(23 downto 16) <= (others => '0');
    if wr = '1' then
      case addr is
        when x"10" =>
          axis(0).reg_ratelatch(7 downto 0) <= din;
        when x"11" =>
          axis(0).reg_ratelatch(15 downto 8) <= din;
        when x"12" =>
          axis(0).reg_rate <= din & axis(0).reg_ratelatch(15 downto 0);
        when others =>
          null;
      end case;
    end if;
  end if;
end process;

--
-- Rate generator 1 rate register
--

reg_ax1_rate_p: process(clk100m, reset)
begin
  if reset = '1' then
    axis(1).reg_rate <= (others => '0');
    axis(1).reg_ratelatch <= (others => '0');
  elsif rising_edge(clk100m) then
    axis(1).reg_ratelatch(23 downto 16) <= (others => '0');
    if wr = '1' then
      case addr is
        when x"13" =>
          axis(1).reg_ratelatch(7 downto 0) <= din;
        when x"14" =>
          axis(1).reg_ratelatch(15 downto 8) <= din;
        when x"15" =>
          axis(1).reg_rate <= din & axis(1).reg_ratelatch(15 downto 0);
        when others =>
          null;
      end case;
    end if;
  end if;
end process;

--
-- Rate generator 2 rate register
--

reg_ax2_rate_p: process(clk100m, reset)
begin
  if reset = '1' then
    axis(2).reg_rate <= (others => '0');
    axis(2).reg_ratelatch <= (others => '0');
  elsif rising_edge(clk100m) then
    axis(2).reg_ratelatch(23 downto 16) <= (others => '0');
    if wr = '1' then
      case addr is
        when x"16" =>
          axis(2).reg_ratelatch(7 downto 0) <= din;
        when x"17" =>
          axis(2).reg_ratelatch(15 downto 8) <= din;
        when x"18" =>
          axis(2).reg_rate <= din & axis(2).reg_ratelatch(15 downto 0);
        when others =>
          null;
      end case;
    end if;
  end if;
end process;

--
-- Rate generator control register
--

reg_rgctrl_p: process(clk100m, reset)
begin
  if reset = '1' then
    reg_rgctrl <= (others => '0');
  elsif rising_edge(clk100m) then
    if addr = x"1C" then
      if wr = '1' then
        reg_rgctrl <= din;
      end if;
    end if;
  end if;
end process;

--
-- Rate generator control register
--

reg_dout_p: process(clk100m, reset)
begin
  if reset = '1' then
    reg_dout <= (others => '0');
  elsif rising_edge(clk100m) then
    if addr = x"1F" then
      if wr = '1' then
        reg_dout <= din;
      end if;
    end if;
  end if;
end process;

end rtl;