--
-- EPP interface
--
-- 2010-09-27 by JTA
--
-- Updates:
--
-- 2010-10-16 by JTA -> - Added delay in bus transactions to avoid timing problems
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.epp_if_pkg.all;
use work.isync_pkg.all;

entity epp_if is
generic
(
  WAITCOUNT         : natural
);
port
(
  clk               : in std_logic;
  reset             : in std_logic;
  --
  epp_data          : inout std_logic_vector(7 downto 0);
  epp_wait          : out std_logic;
  epp_dir_in        : in std_logic;
  epp_astrobeb_in   : in std_logic;
  epp_dstrobeb_in   : in std_logic;
  epp_reset_in      : in std_logic;
  --
  addr              : out std_logic_vector(7 downto 0);
  dout              : out std_logic_vector(7 downto 0);
  din               : in std_logic_vector(7 downto 0);
  rd                : out std_logic;
  wr                : out std_logic;
  waitrequest       : in std_logic;
  reset_out         : out std_logic;
  addr_autoinc_ena  : in std_logic
);
end epp_if;

architecture RTL of epp_if is

type EPP_BUSSTATE_T is
(
  BUSS_IDLE,
  BUSS_DATA_WR,
  BUSS_DATA_RD,
  BUSS_WAITSTATE,
  BUSS_WAITEND
);

type EPP_BUSOP_T is
(
  BUSOP_NONE,
  BUSOP_ADDR_WR,
  BUSOP_ADDR_RD,
  BUSOP_DATA_WR,
  BUSOP_DATA_RD
);

signal reg_addr : unsigned(7 downto 0);
signal epp_busstate : EPP_BUSSTATE_T := BUSS_IDLE;
signal epp_busop : EPP_BUSOP_T := BUSOP_NONE;
signal epp_oen : std_logic := '0';
signal epp_dout : std_logic_vector(7 downto 0);
signal epp_din : std_logic_vector(7 downto 0);
signal epp_dir : std_logic;
signal epp_dstrobeb : std_logic;
signal epp_astrobeb : std_logic;
signal epp_ctrl_in : std_logic_vector(3 downto 0);
signal epp_resetb : std_logic;
signal waitcntr : natural range 0 to WAITCOUNT;

begin

--
-- Parallel port synchronizers
--

epp_ctrl_in <= 
(
  0 => epp_dir_in,
  1 => epp_dstrobeb_in,
  2 => epp_astrobeb_in,
  3 => epp_reset_in
);

ppc_isync: isync_slv
generic map
(
  SYNC_STAGES => 2,
  WIDTH => epp_ctrl_in'length
)
port map
(
  clk => clk, 
  reset => reset, 
  sin =>  epp_ctrl_in, 
  sout(0) => epp_dir,
  sout(1) => epp_dstrobeb,
  sout(2) => epp_astrobeb,
  sout(3) => epp_resetb
);

ppd_isync: isync_slv
generic map
(
  SYNC_STAGES => 2,
  WIDTH => epp_data'length
)
port map
(
  clk => clk, 
  reset => reset, 
  sin =>  epp_data, 
  sout => epp_din
);

epp_data <= epp_dout when epp_oen = '1' else (others => 'Z');
addr <= std_logic_vector(reg_addr);
reset_out <= epp_resetb;

--
-- EPP signalling handling
--

process(clk, reset)
begin
  if reset = '1' then
    reg_addr <= (others => '0');
    epp_busstate <= BUSS_IDLE;
    epp_busop <= BUSOP_NONE;
    epp_wait <= '0';
    epp_oen <= '0';
    epp_dout <= (others => '0');
    rd <= '0';
    wr <= '0';
    dout <= (others => '0');
  elsif rising_edge(clk) then
    case epp_busstate is
      when BUSS_IDLE =>
        epp_wait <= '0';
        epp_oen <= '0';
        waitcntr <= WAITCOUNT;
        epp_busop <= BUSOP_NONE;
        
        if epp_astrobeb = '0' and epp_dstrobeb = '1' then
          if epp_dir = '0' then
            epp_busstate <= BUSS_WAITSTATE;            
            epp_busop <= BUSOP_ADDR_WR;
          else
            epp_dout <= std_logic_vector(reg_addr);
            epp_oen <= '1';
            epp_busstate <= BUSS_WAITSTATE;
            epp_busop <= BUSOP_ADDR_RD;
          end if;
        elsif epp_astrobeb = '1' and epp_dstrobeb = '0' then
          if epp_dir = '0' then
            epp_busstate <= BUSS_WAITSTATE;
            epp_busop <= BUSOP_DATA_WR;
          else
            epp_busstate <= BUSS_DATA_RD;
            epp_busop <= BUSOP_DATA_RD;
            rd <= '1';
          end if;
        end if;

      when BUSS_DATA_RD =>
        if waitrequest = '0' then
          if addr_autoinc_ena ='1' then
            reg_addr <= reg_addr + 1;
          end if;
          epp_dout <= din;
          epp_oen <= '1';
          rd <= '0';
          epp_busstate <= BUSS_WAITSTATE;
        end if;

      when BUSS_DATA_WR =>
        if waitrequest = '0' then
          if addr_autoinc_ena = '1' then
            reg_addr <= reg_addr + 1;
          end if;
          epp_busstate <= BUSS_WAITEND;
          wr <= '0';
          epp_wait <= '1';
        end if;

      when BUSS_WAITSTATE =>
        if waitcntr > 0 then
          waitcntr <= waitcntr - 1;
        else
          case epp_busop is
            when BUSOP_ADDR_WR =>
              reg_addr <= unsigned(epp_din);
              epp_busstate <= BUSS_WAITEND;
              epp_wait <= '1';
            when BUSOP_DATA_WR =>
              epp_busstate <= BUSS_DATA_WR;
              dout <= epp_din;
              wr <= '1';
            when others =>
              epp_busstate <= BUSS_WAITEND;
              epp_wait <= '1';            
          end case;
        end if;

      when BUSS_WAITEND =>
        if epp_astrobeb = '1' and epp_dstrobeb = '1' then
          epp_busstate <= BUSS_IDLE;
          epp_oen <= '0';
          epp_wait <= '0';
        end if;

      when others => null;
    end case;
  end if;
end process;

end RTL;
