--
-- Step & dir generator from up & down-pulses
--
-- 2010-01-22 by JTA
--

library ieee;
use ieee.std_logic_1164.all;

entity stepdir_gen is
generic
(
  STEP_WIDTH : integer;
  DIR_SETUP : integer
);
port
(
  clk : in std_logic;
  reset : in std_logic;
  
  fwd_in : in std_logic;
  rev_in : in std_logic;
  enable : in std_logic;
  
  step : out std_logic;
  dir : out std_logic
);
end;

architecture rtl of stepdir_gen is

type SDGEN_STATE_T is (SDG_WAIT, SDG_PH0, SDG_PH1);
type STEP_T is (STEP_FWD, STEP_REV, STEP_NONE);

signal sdg_state : SDGEN_STATE_T := SDG_WAIT;
signal next_dir : std_logic := '0';
signal do_step : boolean := false;
signal step_taken : boolean := false;

signal stepwaitcounter : integer range 0 to (STEP_WIDTH - 1) := STEP_WIDTH - 1;
signal dirsetupcounter : integer range 0 to (DIR_SETUP - 1) := DIR_SETUP - 1;
signal start_stepdelay : boolean := false;
signal stepdelay_expired : boolean := false;
signal start_dirsetupdelay : boolean := false;
signal dirsetupdelay_expired : boolean := false;

begin

process(clk, reset)
begin
  if reset = '1' then
    do_step <= false;
  elsif rising_edge(clk) then
    if (not do_step) and (enable = '1') then
      if fwd_in = '1' and rev_in = '0' then 
        do_step <= true;        
        next_dir <= '0';
      elsif fwd_in = '0' and rev_in = '1' then
        do_step <= true;
        next_dir <= '1';
      end if;
    else
      if step_taken then
        do_step <= false;
      end if;
    end if;
  end if;
end process;

--
-- Step width generator
--

process(reset, clk)
begin
  if reset = '1' then
    dirsetupcounter <= DIR_SETUP - 1;
  elsif rising_edge(clk) then
    if start_dirsetupdelay then
      dirsetupcounter <= DIR_SETUP - 1;
    elsif dirsetupcounter > 0 then
      dirsetupcounter <= dirsetupcounter - 1;
    end if;     
  end if;   
end process;

dirsetupdelay_expired <= true when dirsetupcounter = 0 and not start_dirsetupdelay else false;

--
-- Step width generator
--

process(reset, clk)
begin
  if reset = '1' then
    stepwaitcounter <= STEP_WIDTH - 1;
  elsif rising_edge(clk) then
    if start_stepdelay then
      stepwaitcounter <= STEP_WIDTH - 1;
    elsif stepwaitcounter > 0 then
      stepwaitcounter <= stepwaitcounter - 1;
    end if;     
  end if;   
end process;

stepdelay_expired <= true when stepwaitcounter = 0 and not start_stepdelay else false;

process(reset, clk)
begin
  if reset = '1' then
    sdg_state <= SDG_WAIT;
    dir <= '0';
    step <= '0';
  elsif rising_edge(clk) then   
    case sdg_state is
      when SDG_WAIT =>
        if do_step then
          step_taken <= true;
          dir <= next_dir;
          start_dirsetupdelay <= true;
          sdg_state <= SDG_PH0;
        end if;
          
      when SDG_PH0 =>
        step_taken <= false;
        start_dirsetupdelay <= false;
        if dirsetupdelay_expired then
          step <= '1';
          start_stepdelay <= true;
          sdg_state <= SDG_PH1;
        end if;
          
      when SDG_PH1 =>
        start_stepdelay <= false;       
        if stepdelay_expired then
          step <= '0';
          if do_step then
            step_taken <= true;
            dir <= next_dir;
            sdg_state <= SDG_PH0;
            start_dirsetupdelay <= true;
          else
            sdg_state <= SDG_WAIT;
          end if;
        end if;

      when others =>
    end case; 
  end if;
end process;  
  
end rtl;
